package Lire::AsciiDlf::GroupOp;

use strict;

use base qw/ Lire::AsciiDlf::ReportOperator /;

use Lire::GroupOp;
use Lire::DataTypes qw/ format_numeric_type /;
use Carp;

use vars qw/ $VERSION /;

BEGIN {
    ($VERSION)	= '$Revision: 1.39 $' =~ m!Revision: ([.\d]+)!;
}

#------------------------------------------------------------------------
# Method init_report( $period_start, $period_end)
#
# Method required by Lire::AsciiDlf::ReportOperator
#
# Default implementation should be chained up to
# initialize properly the missing cases tracker.
sub init_report {
    my ( $self, $period_start, $period_end ) = @_;

    $self->{_missing_cases} = {};

    $self;
}

#------------------------------------------------------------------------
# Method init_merge( $period_start, $period_end )
#
# Method required by Lire::AsciiDlf::ReportOperator
#
# Default implementation does nothing
sub init_merge {
    my ( $self, $period_start, $period_end ) = @_;
}

#------------------------------------------------------------------------
# Method end_report()
#
# Method required by Lire::AsciiDlf::ReportOperator
#
# Default implementation does nothing
sub end_report {}

#------------------------------------------------------------------------
# Method end_merge()
#
# Method required by Lire::AsciiDlf::ReportOperator
#
# Default implementation does nothing
sub end_merge {}

#------------------------------------------------------------------------
# Method init_group_data()
#
# Method required by Lire::AsciiDlf::ReportOperator
#
# OPTIMIZATION OF THE DATA STRUCTURE
#.
# Scalar ref takes less place then array refs which takes less spec
# then hash ref. We use only scalar or array refs.
#
# As a side effects, we also get a little speed optimization :
# perl -MBenchmark=cmpthese -e '
#   my $dummy = 0;
#   my $scalar = \$dummy;
#   my $array = [0]; 
#   my $hash = { value => 0};
#   cmpthese( 500000, {
#		SCALAR => sub { $$scalar++ },
#		ARRAY  => sub { $array->[0]++ },
#		HASH   => sub { $hash->{value}++ },
#	     });
# Benchmark: timing 500000 iterations of ARRAY, HASH, SCALAR...
#     ARRAY:  0 wallclock secs ( 0.57 usr +  0.00 sys =  0.57 CPU) @ 877192.98/s (n=500000)
#     HASH:  1 wallclock secs ( 0.67 usr +  0.00 sys =  0.67 CPU) @ 746268.66/s (n=500000)
#    SCALAR:  1 wallclock secs ( 0.41 usr +  0.00 sys =  0.41 CPU) @ 1219512.20/s (n=500000)
#             Rate   HASH  ARRAY SCALAR
# HASH    746269/s     --   -15%   -39%
# ARRAY   877193/s    18%     --   -28%
# SCALAR 1219512/s    63%    39%     --
#
# Make default implementation uses a scalar reference
sub init_group_data {
    my $scalar = 0;
    return \$scalar;
}

#------------------------------------------------------------------------
# Method end_group_data( $data )
#
# Method required by Lire::AsciiDlf::ReportOperator
#
# Default implementation does nothing
sub end_group_data {}

#------------------------------------------------------------------------
# Method add_entry_value( $entry, $data )
#
# Method required by Lire::AsciiDlf::ReportOperator
#
# Default implementation formats the value using
# format_content_value() with the value held in the scalar reference.
#
# Subclasses will need to override it if they override init_group_data()
sub add_entry_value {
    my ( $self, $entry, $data ) = @_;

    my $content = $self->format_content_value( $$data );
    $entry->add_value( content => $content,
                       value   => $$data,
                       missing_cases => $self->missing_cases( $data ),
                     );
}

#------------------------------------------------------------------------
# Method format_content_value( $value )
#
# Convenient method for add_entry_value() implementation which returns
# the value formated using format_numeric_type().

# The type used is either the value in the C<field_type> attribute or
# C<number>.
#
# Subclasses might want to set C<field_type> appropriately.
sub format_content_value {
    my ( $self, $value ) = @_;

    return format_numeric_type( $value, $self->{field_type} || "number" );
}

#------------------------------------------------------------------------
# set_summary_value( $group, $value )
#
# Used by Lire::AsciiDlf::Aggregator
sub set_summary_value {
    croak "Unimplemented set_summary_value() in ", ref $_[0];
}

#------------------------------------------------------------------------
# report_missing_case( $data )
#
# This method should be used by subclass that encounters 
# undefined value. This will increase the missing count associated
# with $data by one.
sub report_missing_case {
    my ( $self, $data ) = @_;

    croak "missing 'data' parameter"
      unless defined $data;

    $self->{_missing_cases}{$data} = 0
      unless defined $self->{_missing_cases}{$data};
    $self->{_missing_cases}{$data}++;
}

#------------------------------------------------------------------------
# missing_cases( $data )
#
# This method returns the number of missing-cases encountered
# while processing $data.
sub missing_cases {
    my ( $self, $data );

    return defined $self->{_missing_cases}{$data} ?
      $self->{_missing_cases}{$data} : 0;
}

package Lire::AsciiDlf::GroupOp::Count;

use base qw/ Lire::AsciiDlf::GroupOp Lire::GroupOp::Count /;

use Lire::DataTypes qw/ format_numeric_type /;
use Lire::Utils qw/ratio100/;
use Carp;

#------------------------------------------------------------------------
# Method init_report( $period_start, $period_end )
#
# Method required by Lire::AsciiDlf::ReportOperator
sub init_report {
    my ($self, $period_start, $period_end) = @_;

    if ( $self->fields() ) {
	$self->{key_maker} =
	  $self->report_spec->schema->make_key_access_func( @{$self->fields} );
    }

    $self;
}

#------------------------------------------------------------------------
# Method init_group_data()
#
# Method required by Lire::AsciiDlf::ReportOperator
sub init_group_data {
    my ( $self ) = @_;

    if ( $self->{key_maker} ) {
        return [ 0, {} ];
    } else {
	my $scalar = 0;
	return \$scalar;
    }
}

sub is_missing_case {
    my ( $self, $dlf ) = @_;

    if ( $self->{key_maker} ) {
        return ! defined $self->{key_maker}->( $dlf );
    } else {
        return 0;
    }
}

#------------------------------------------------------------------------
# Method update_group_data( $dlf, $data )
#
# Method required by Lire::AsciiDlf::ReportOperator
sub update_group_data {
    my ( $self, $dlf, $data ) = @_;

    return $self->report_missing_case( $data )
      if $self->is_missing_case( $dlf );

    if ( $self->{key_maker} ) {
	my $key = $self->{key_maker}->( $dlf );
        if ( ! exists $data->[1]{$key} ) {
	    $data->[1]{$key} = 1;
            $data->[0]++;
        }
    } else {
	$$data++;
    }
}

#------------------------------------------------------------------------
# Method merge_group_data( $value, $data )
#
# Method required by Lire::AsciiDlf::ReportOperator
sub merge_group_data {
    my ( $self, $value, $data ) = @_;

    # FIXME: We should add a note when a key_maker was used. For
    # example, there is no guarentee that the sum of the number of
    # hosts that visited during two reports is the same than the
    # number of hosts calculated over the period spanned by the two
    # reports.

    # To merge two counts, we simply add them
    $$data += $value->{value};
}

#------------------------------------------------------------------------
# Method end_group_data( $data )
#
# Method required by Lire::AsciiDlf::ReportOperator
sub end_group_data {
    my ( $self, $data ) = @_;

    $data->[1] = undef
      if $self->{key_maker};
}

#------------------------------------------------------------------------
# Method set_summary_value( $group, $data )
#
# Method required by Lire::AsciiDlf::GroupOp.
# We may need to format the value according to the ratio parameter
sub set_summary_value {
    my ( $self, $group, $data ) = @_;

    my $type  = $self->{field_type} || "number";
    my $value = $self->{key_maker} ? $data->[0] : $$data;
    my $ratio_spec = $self->ratio;
    my $content;
    if ($ratio_spec eq 'none' ) {
	$content = format_numeric_type( $value, $type );
    } elsif ( ! $group->parent_entry ) {
	# We are in the top-level container
	$content = "100.0";
    } else {
	my $container = $ratio_spec eq 'table' ? $group->parent_entry->subreport :
	  $group->parent_entry->group;
	my $total = $container->get_summary_value( $self->name );
	$content = ratio100( $value, $total->{value} );
    }

    $group->set_summary_value( $self->name,
                               content => $content,
                               value => $value,
                               missing_cases => $self->missing_cases( $data ),
                              );
}

#------------------------------------------------------------------------
# Method add_entry_value( $entry, $data )
#
# Method required by Lire::AsciiDlf::ReportOperator
#
# Data structure may vary and we may have to handle ratio
# information
sub add_entry_value {
    my ( $self, $entry, $data ) = @_;

    my $value = $self->{key_maker} ? $data->[0] : $$data;
    my $ratio_spec = $self->ratio;
    if ( $ratio_spec eq 'none' ) {
        $entry->add_value( content => $self->format_content_value( $value ),
                           value   => $value,
                           missing_cases => $self->missing_cases( $data ),
                         );
	return;
    }

    my $group;
    if ($ratio_spec eq 'table' ) {
	$group = $entry->subreport;
    } else {
	$group = $entry->group;
    }

    my $total = $group->get_summary_value( $self->name );
    my $ratio = ratio100( $value, $total->{value} );
    $entry->add_value( content => $ratio,
                       value => $value,
                       missing_cases => $self->missing_cases( $data ),
                     );
}

package Lire::AsciiDlf::GroupOp::SimpleStat;

use base qw/ Lire::AsciiDlf::GroupOp /;

use Lire::DataTypes qw/ format_numeric_type /;

#------------------------------------------------------------------------
# Method init_report( $period_start, $period_end )
#
# Method required by Lire::AsciiDlf::ReportOperator
sub init_report {
    my ($self, $period_start, $period_end ) = @_;

    my $field = $self->report_spec->schema->field( $self->field );
    $self->{field_idx}  = $field->pos;
    $self->{field_type} = $field->type;

    $self;
}

#------------------------------------------------------------------------
# Method init_merge( $period_start, $period_end)
#
# Method required by Lire::AsciiDlf::ReportOperator
sub init_merge {
    my ($self, $period_start, $period_end ) = @_;

    my $field = $self->{report_spec}->schema()->field( $self->field );
    $self->{field_type} = $field->type;

    $self;
}

#------------------------------------------------------------------------
# Method set_summary_value( $group, $data )
#
# Method required by Lire::AsciiDlf::GroupOp
sub set_summary_value {
    my ( $self, $group, $data ) = @_;

    my $value = $$data;
    my $type  = $self->{field_type} || "number";
    my $content   = format_numeric_type( $value, $type );

    $group->set_summary_value( $self->name,
                               content => $content,
                               value => $value,
                               missing_cases => $self->missing_cases( $data ),
                              );
}

package Lire::AsciiDlf::GroupOp::Sum;

use base qw/ Lire::AsciiDlf::GroupOp::SimpleStat Lire::GroupOp::Sum /;

use Lire::DataTypes qw/format_numeric_type/;
use Lire::Utils qw/ratio100/;

#------------------------------------------------------------------------
# Method init_report( $period_start, $period_end )
#
# Method required by Lire::AsciiDlf::ReportOperator.
#
# We need to cache the weight field index if we are computing a weighted
# sum
sub init_report {
    my $self = shift;

    $self->SUPER::init_report( @_ );

    my $weight_f = $self->weight;
    if ( $weight_f ) {
	$self->{weight_field_idx}  = 
	  $self->report_spec->schema->field( $weight_f )->pos;
    } else {
	delete $self->{weight_field_idx};
    }
}

sub is_missing_case {
    my ( $self, $dlf ) = @_;

    if ( $self->{weight_field_idx} ) {
        return ! defined $dlf->[$self->{weight_field_idx}] ||
          ! defined $dlf->[$self->{field_idx}];
    } else {
        return ! defined $dlf->[$self->{field_idx}];
    }
}

#------------------------------------------------------------------------
# Method update_group_data( $dlf, $data )
#
# Method required by Lire::AsciiDlf::ReportOperator
sub update_group_data {
    my ( $self, $dlf, $data ) = @_;

    return $self->report_missing_case( $data )
      if $self->is_missing_case( $dlf );

    if ( $self->{weight_field_idx} ) {
	$$data +=
	  $dlf->[$self->{field_idx}] * $dlf->[$self->{weight_field_idx}];
    } else {
	$$data += $dlf->[$self->{field_idx}];
    }
}

#------------------------------------------------------------------------
# Method merge_group_data( $value, $data )
#
# Method required by Lire::AsciiDlf::ReportOperator
sub merge_group_data {
    my ( $self, $value, $data ) = @_;

    # To merge two sums, we simply add them
    $$data += $value->{value};
}

#------------------------------------------------------------------------
# Method set_summary_value( $group, $data )
#
# Method required by Lire::AsciiDlf::GroupOp.
# We may need to format the value according to the ratio parameter
sub set_summary_value {
    my ( $self, $group, $data ) = @_;

    my $value = $$data;
    my $type	 = $self->{field_type} || "number";
    my $ratio_spec = $self->ratio;
    my $content;
    if ($ratio_spec eq 'none' ) {
	$content = format_numeric_type( $value, $type );
    } elsif ( ! $group->parent_entry ) {
	# We are in the top-level container
	$content = "100.0";
    } else {
	my $container = $ratio_spec eq 'table' ? $group->parent_entry->subreport :
	  $group->parent_entry->group;
	my $total = $container->get_summary_value( $self->name );
	$content = ratio100( $value, $total->{value} );
    }

    $group->set_summary_value( $self->name,
                                content => $content,
                                value   => $value,
                                missing_cases => $self->missing_cases( $data ),
                              );

}

#------------------------------------------------------------------------
# Method add_entry_value( $entry, $data )
#
# Method required by Lire::AsciiDlf::ReportOperator
#
# We may need to print the sum as a ratio
sub add_entry_value {
    my ( $self, $entry, $data ) = @_;

    my $value = $$data;
    my $ratio_spec = $self->ratio;
    if ( $ratio_spec eq 'none' ) {
	$self->SUPER::add_entry_value( $entry, $data );
	return;
    }

    my $group;
    if ($ratio_spec eq 'table' ) {
	$group = $entry->subreport;
    } else {
	$group = $entry->group;
    }

    my $total = $group->get_summary_value( $self->name );
    my $ratio = ratio100( $value, $total->{value} );
    $entry->add_value( content => $ratio,
                       value   => $value,
                       missing_cases => $self->missing_cases( $data, ),
                     );
}

package Lire::AsciiDlf::GroupOp::Avg;

use base qw/ Lire::AsciiDlf::GroupOp Lire::GroupOp::Avg /;

use Lire::AsciiDlf::ReportOperator qw/group_data_value/;
use Lire::DataTypes qw/ is_numeric_type format_numeric_type/;

use Carp;

#------------------------------------------------------------------------
# Method init_report( $period_start, $period_end )
#
# Method required by Lire::AsciiDlf::ReportOperator
#
# We implement avg using two other operators either one sum and one count
# or two counts.
sub init_report {
    my $self = shift;

    $self->create_avg_ops;
    $self->{sum_op}->init_report( @_ );
    $self->{n_op}->init_report( @_ );

    $self;
}

#------------------------------------------------------------------------
# Method init_merge( $period_start, $period_end )
#
# Method required by Lire::AsciiDlf::ReportOperator
sub init_merge {
    my ( $self, $period_start, $period_end ) = @_;

    $self->create_avg_ops;
    $self->{sum_op}->init_merge( $period_start, $period_end );
    $self->{n_op}->init_merge( $period_start, $period_end );

    $self
}

#------------------------------------------------------------------------
# Method create_avg_ops()
#
# Initialize n_op and sum_op
sub create_avg_ops {
    my ($self) = @_;
    my %common = (
		  parent => $self->parent,
		  report_spec => $self->report_spec,
		 );
    if ( $self->field ) {
	my $f = $self->report_spec->schema->field( $self->field );
	if ( is_numeric_type( $f->type ) ) {
	    # Use a sum
	    $self->{sum_op} =
	      new Lire::AsciiDlf::GroupOp::Sum( %common,
						name => $self->name . ".sum",
						field => $self->field,
						weight => $self->weight,
					      );
	} else {
	    # Use a counter
	    $self->{sum_op} =
	      new Lire::AsciiDlf::GroupOp::Count( %common,
						  fields => [ $self->field ],
						  name => $self->name . ".sum",
						);
	}
	$self->{field_type} = $f->type;
    } else {
	# Use a simple count operation
	$self->{sum_op} = 
	  new Lire::AsciiDlf::GroupOp::Count( %common,
					      name => $self->name . ".sum",
					    );
    }

    # Special case for a numerical field. This enables
    # to compute arbitrary ratio.
    # FIXME: A new operator ratio operator should be defined
    if ( $self->by_fields && @{$self->by_fields} == 1 &&
         is_numeric_type( $self->report_spec->schema->field( $self->by_fields->[0])->type))
     {
         $self->{n_op} =
           new Lire::AsciiDlf::GroupOp::Sum( %common,
                                             field => $self->by_fields->[0],
                                             name => $self->name . ".n",
                                           );
     } else {
         $self->{n_op} =
           new Lire::AsciiDlf::GroupOp::Count( %common,
                                               fields => $self->by_fields,
                                               name => $self->name . ".n",
                                             );
     }
}

#------------------------------------------------------------------------
# Method dlf_sort_fields()
#
# Method required by Lire::AsciiDlf::ReportOperator
#
# Returns whatever is required by our sum or n operator. If both have
# requirements, we have to check that they are compatible.
sub dlf_sort_fields {
    my ( $self ) = @_;

    my @sum_f = $self->{sum_op}->dlf_sort_fields;
    my @n_f   = $self->{n_op}->dlf_sort_fields;

    return @sum_f if @sum_f && !@n_f;
    return @n_f if @n_f && !@sum_f;

    # We may have a conflict.
    my $last = @sum_f > @n_f ? @sum_f : @n_f;
    foreach (my $i=0; $i < $last; $i++ ) {
	die "conflicting sort requirements in average implementation: ",
	  join ( ",", @sum_f ), " != ", join( ",", @n_f), "\n"
	    if $sum_f[$i] ne $n_f[$i];
    }

    # Return the longest
    return @sum_f > @n_f ? @sum_f : @n_f;
}

#------------------------------------------------------------------------
# Method init_group_data()
#
# Method required by Lire::AsciiDlf::ReportOperator
sub init_group_data {
    my ( $self ) = @_;

    # Result is held in the first element
    [ "NaN", $self->{sum_op}->init_group_data, $self->{n_op}->init_group_data ];
}

#------------------------------------------------------------------------
# Method update_group_data( $dlf, $data )
#
# Method required by Lire::AsciiDlf::ReportOperator
sub update_group_data {
    my ( $self, $dlf, $data ) = @_;

    return $self->report_missing_case( $data )
      if $self->{sum_op}->is_missing_case( $dlf ) ||
        $self->{n_op}->is_missing_case( $dlf );

    $self->{sum_op}->update_group_data( $dlf, $data->[1] );
    $self->{n_op}->update_group_data( $dlf, $data->[2] );
}

#------------------------------------------------------------------------
# Method merge_group_data( $value, $data )
#
# Method required by Lire::AsciiDlf::ReportOperator
sub merge_group_data {
    my ( $self, $value, $data ) = @_;

    $self->{sum_op}->merge_group_data( {value => $value->{total}},
				       $data->[1] );
    $self->{n_op}->merge_group_data( {value => $value->{n}},
				     $data->[2] );
}

#------------------------------------------------------------------------
# Method end_group_data( $data )
#
# Method required by Lire::AsciiDlf::ReportOperator
sub end_group_data {
    my ( $self, $data ) = @_;

    $self->{sum_op}->end_group_data( $data->[1] );
    $self->{n_op}->end_group_data( $data->[2] );

    my $sum = group_data_value( $data->[1] );
    my $n = group_data_value( $data->[2] );
    if ( $n ) {
	$data->[0] = sprintf "%.2f", $sum / $n;
    }
}

#------------------------------------------------------------------------
# Method end_report()
#
# Method required by Lire::AsciiDlf::ReportOperator
sub end_report {
    my ( $self ) = @_;

    $self->{sum_op}->end_report;
    $self->{n_op}->end_report;
}

#------------------------------------------------------------------------
# Method end_merge()
#
# Method required by Lire::AsciiDlf::ReportOperator
sub end_merge {
    my ( $self ) = @_;

    $self->{sum_op}->end_merge;
    $self->{n_op}->end_merge;
}

#------------------------------------------------------------------------
# Method add_entry_value( $entry, $data )
#
# Method required by Lire::AsciiDlf::ReportOperator
sub add_entry_value {
    my ( $self, $entry, $data ) = @_;

    my $type	 = $self->{field_type} || "number";
    my $content  = format_numeric_type( $data->[0], $type );

    $entry->add_value( content => $content,
                       value   => $data->[0], 
                       total   => group_data_value($data->[1]),
                       n       => group_data_value( $data->[2]),
                       missing_cases => $self->missing_cases( $data ),
                     );
}

#------------------------------------------------------------------------
# Method set_summary_value( $group, $data )
#
# Method required by Lire::AsciiDlf::GroupOp
sub set_summary_value {
    my ( $self, $group, $data ) = @_;

    my $type	 = $self->{field_type} || "number";
    my $value = $data->[0];
    my $content   = format_numeric_type( $value, $type );

    $group->set_summary_value( $self->name,
                                content => $content,
                                value => $value,
                                total => group_data_value($data->[1]),
                                n => group_data_value( $data->[2] ),
                                missing_cases => $self->missing_cases( $data ),
                              );
}

package Lire::AsciiDlf::GroupOp::Min;

use base qw/ Lire::AsciiDlf::GroupOp::SimpleStat Lire::GroupOp::Min /;

#------------------------------------------------------------------------
# Method init_group_data()
#
# Method required by Lire::AsciiDlf::ReportOperator
sub init_group_data {
    my $scalar = undef;
    return \$scalar;
}

#------------------------------------------------------------------------
# Method update_group_data( $dlf, $data )
#
# Method required by Lire::AsciiDlf::ReportOperator
sub update_group_data {
    my ( $self, $dlf, $data ) = @_;

    my $value = $dlf->[$self->{field_idx}];
    return $self->report_missing_case( $data)
      unless defined $value;

    $$data = $value
      unless ( defined $$data );

    $$data = $value
      if $value < $$data;
}

#------------------------------------------------------------------------
# Method merge_group_data( $dlf, $data )
#
# Method required by Lire::AsciiDlf::ReportOperator
sub merge_group_data {
    my ( $self, $value, $data ) = @_;

    return if lc $value->{value} eq 'nan';

    $$data = $value->{value}
      unless ( defined $$data );

    # To merge two mins, we keep the lowest
    $$data = $value->{value}
      if $value->{value} < $$data;
}

#------------------------------------------------------------------------
# Method end_group_data( $data )
#
# Method required by Lire::AsciiDlf::ReportOperator
sub end_group_data {
    my ( $self, $data ) = @_;

    $$data = "NaN" unless defined $$data;
}

package Lire::AsciiDlf::GroupOp::Max;

use base qw/ Lire::AsciiDlf::GroupOp::SimpleStat Lire::GroupOp::Max  /;

#------------------------------------------------------------------------
# Method init_group_data()
#
# Method required by Lire::AsciiDlf::ReportOperator
sub init_group_data {
    my $scalar = undef;
    return \$scalar;
}

#------------------------------------------------------------------------
# Method update_group_data( $dlf, $data )
#
# Method required by Lire::AsciiDlf::ReportOperator
sub update_group_data {
    my ( $self, $dlf, $data ) = @_;

    my $value = $dlf->[$self->{field_idx}];
    return $self->report_missing_case( $data )
      unless defined $value;

    $$data = $value
      unless ( defined $$data );

    $$data = $value
      if $value > $$data;
}

#------------------------------------------------------------------------
# Method merge_group_data( $dlf, $data )
#
# Method required by Lire::AsciiDlf::ReportOperator
sub merge_group_data {
    my ( $self, $value, $data ) = @_;

    return if lc $value->{value} eq 'nan';

    $$data = $value->{value}
      unless ( defined $$data );

    # To merge two max, we keep the highest
    $$data = $value->{value}
      if $value->{value} > $$data;
}

#------------------------------------------------------------------------
# Method end_group_data( $data )
#
# Method required by Lire::AsciiDlf::ReportOperator
sub end_group_data {
    my ( $self, $data ) = @_;

    $$data = "NaN" unless defined $$data;
}

package Lire::AsciiDlf::GroupOp::First;

use base qw/ Lire::AsciiDlf::GroupOp::SimpleStat Lire::GroupOp::First /;

use Lire::DataTypes qw/ is_numeric_type format_numeric_type /;

use Carp;

#------------------------------------------------------------------------
# Method init_report( $period_start, $period_end )
#
# Method required by Lire::AsciiDlf::ReportOperator
sub init_report {
    my $self = shift;

    $self->SUPER::init_report( @_ );

    my $sort_fields = $self->sort_fields;
    unless ( $sort_fields && @$sort_fields ) {
	$sort_fields = [ $self->report_spec->schema->timestamp_field->name ];
    }

    $self->{sort_func} =
      $self->report_spec->schema->make_dlf_sort_func( @$sort_fields );

    $self->init_common;

    $self;
}

#------------------------------------------------------------------------
# Method init_merge( $period_start, $period_end )
#
# Method required by Lire::AsciiDlf::ReportOperator
sub init_merge {
    my ($self, $period_start, $period_end ) = @_;

    $self->SUPER::init_merge( $period_start, $period_end );

    my $sort_fields = $self->sort_fields;
    unless ( $sort_fields && @$sort_fields ) {
	$sort_fields = [ $self->report_spec->schema->timestamp_field->name ];
    }

    my @cmp = ();
    my $i = 0;
    foreach my $f ( @$sort_fields ) {
	my $type = $self->report_spec->schema->field( $f )->type;
	my $cmp = is_numeric_type( $type ) ? '<=>' : 'cmp';

        push @cmp, "\$_[0][$i] $cmp \$_[1][$i]";
 	$i++;
    }

    my $sort_code = "sub { " . join( " && ", @cmp ) . " }";
    $self->{sort_func} = eval $sort_code;
    die "failed to compile sort function ($sort_code): $@" if $@;

    $self->init_common;

    $self;
}

#------------------------------------------------------------------------
# Method init_common()
#
# Initialization common to both init_report() and init_merge()
sub init_common {
    my ( $self) = @_;

    my $sort_fields = $self->sort_fields;
    unless ( $sort_fields && @$sort_fields ) {
	$sort_fields = [ $self->report_spec->schema->timestamp_field->name ];
    }

    my @i = ();
    foreach my $f ( @$sort_fields ) {
	push @i, $self->report_spec->schema->field( $f )->pos;
    }

    # Extracts the sort values to save them in the total attribute
    $self->{total_func} = sub {
        if ( defined $_[0] ) {
            return join( " ", map { $_[0]->[$_] } @i);
        } else {
            return ""
        }
    };
}

#------------------------------------------------------------------------
# Method init_group_data()
#
# Method required by Lire::AsciiDlf::ReportOperator
sub init_group_data {
    [];
}

#------------------------------------------------------------------------
# Method update_group_data( $dlf, $data )
#
# Method required by Lire::AsciiDlf::ReportOperator
sub update_group_data {
    my ( $self, $dlf, $data ) = @_;

    return $self->report_missing_case( $data )
      unless defined $dlf->[$self->{field_idx}];

    unless (defined $data->[0]) {
	$data->[0] = $dlf->[$self->{field_idx}];
	$data->[1] = $dlf;
	return;
    }

    # Change the value only if this DLF sorts before the last one
    if ( $self->{sort_func}->( $dlf, $data->[1] ) < 0 ) {
	$data->[0] = $dlf->[$self->{field_idx}];
	$data->[1] = $dlf;
    }
}

#------------------------------------------------------------------------
# Method merge_group_data( $dlf, $data )
#
# Method required by Lire::AsciiDlf::ReportOperator
sub merge_group_data {
    my ( $self, $value, $data ) = @_;

    return unless $value->{total} ne "";

    my $fields = [ split / /, $value->{total} ];

    unless (defined $data->[0]) {
	$data->[0] = $value->{value};
	$data->[1] = $fields;
	return;
    }

    # Change the value only if the fields sorts before the last one
    if ( $self->{sort_func}->( $fields, $data->[1] ) < 0 ) {
	$data->[0] = $value->{value};
	$data->[1] = $fields;
    }
}

#------------------------------------------------------------------------
# Method set_summary_value( $group, $data )
#
# Method required by Lire::AsciiDlf::GroupOp.
# We have to take care of non-numerical values
sub set_summary_value {
    my ( $self, $group, $data ) = @_;

    my $value = $data->[0];
    my $type	 = $self->{field_type} || "number";
    my $total = $self->{total_func}->( $data->[1] );
    my $content;
    if ( is_numeric_type( $self->{field_type} ) ) {
	$content = format_numeric_type( $value, $type );
    } else {
	$content = $value;
    }

    $group->set_summary_value( $self->name,
                                content => $content,
                                value => $value,
                                total => $total,
                                missing_cases => $self->missing_cases( $data ),
                              );
}

#------------------------------------------------------------------------
# Method add_entry_value( $entry, $data )
#
# Method required by Lire::AsciiDlf::ReportOperator
# We have to take care of non-numerical values
sub add_entry_value {
    my ( $self, $entry, $data ) = @_;

    my $value = $data->[0];
    my $total = $self->{total_func}->( $data->[1] );
    my $content;
    if ( is_numeric_type( $self->{field_type} ) ) {
	$content = format_numeric_type( $value, $self->{field_type} );
    } else {
	$content = $value;
    }

    $entry->add_value( content => $content,
                       value => $value,
                       total => $total );
}

package Lire::AsciiDlf::GroupOp::Last;

use base qw/ Lire::AsciiDlf::GroupOp::SimpleStat Lire::GroupOp::Last /;

use Lire::DataTypes qw/ is_numeric_type format_numeric_type /;

use Carp;

#------------------------------------------------------------------------
# Method init_report( $period_start, $period_end )
#
# Method required by Lire::AsciiDlf::ReportOperator
sub init_report {
    my $self = shift;

    $self->SUPER::init_report( @_ );

    my $sort_fields = $self->sort_fields;
    unless ( $sort_fields && @$sort_fields ) {
	$sort_fields = [ $self->report_spec->schema->timestamp_field->name ];
    }

    $self->{sort_func} =
      $self->report_spec->schema->make_dlf_sort_func( @$sort_fields );

    $self->init_common;

    $self;
}

#------------------------------------------------------------------------
# Method init_merge( $period_start, $period_end )
#
# Method required by Lire::AsciiDlf::ReportOperator
sub init_merge {
    my ($self, $period_start, $period_end ) = @_;

    $self->SUPER::init_merge( $period_start, $period_end );

    my $sort_fields = $self->sort_fields;
    unless ( $sort_fields && @$sort_fields ) {
	$sort_fields = [ $self->report_spec->schema->timestamp_field->name ];
    }

    my @cmp = ();
    my $i = 0;
    foreach my $f ( @$sort_fields ) {
	my $type = $self->report_spec->schema->field( $f )->type;
	my $cmp = is_numeric_type( $type ) ? '<=>' : 'cmp';

        push @cmp, "\$_[0][$i] $cmp \$_[1][$i]";
	$i++;
    }

    my $sort_code = "sub { " . join( " && ", @cmp ) . " }";
    $self->{sort_func} = eval $sort_code;
    die "failed to compile sort function ($sort_code): $@" if $@;

    $self->init_common;

    $self;
}

#------------------------------------------------------------------------
# Method init_common()
#
# Initialization common to both init_report() and init_merge()
sub init_common {
    my ( $self) = @_;

    my $sort_fields = $self->sort_fields;
    unless ( $sort_fields && @$sort_fields ) {
	$sort_fields = [ $self->report_spec->schema->timestamp_field->name ];
    }

    my @i = ();
    foreach my $f ( @$sort_fields ) {
	push @i, $self->report_spec->schema->field( $f )->pos;
    }

    # Extracts the sort values to save them in the total attribute
    $self->{total_func} = sub {
        if ( defined $_[0] ) {
            return join( " ", map { $_[0]->[$_] } @i);
        } else {
            return ""
        }
    };
}

#------------------------------------------------------------------------
# Method init_group_data()
#
# Method required by Lire::AsciiDlf::ReportOperator
sub init_group_data {
    [];
}

#------------------------------------------------------------------------
# Method update_group_data( $dlf, $data )
#
# Method required by Lire::AsciiDlf::ReportOperator
sub update_group_data {
    my ( $self, $dlf, $data ) = @_;

    return $self->report_missing_case( $data )
      unless defined $dlf->[$self->{field_idx}];

    unless (defined $data->[0]) {
	$data->[0] = $dlf->[$self->{field_idx}];
	$data->[1] = $dlf;
	return;
    }

    # Change the value only if this DLF sorts after the last one
    if ( $self->{sort_func}->( $dlf, $data->[1] ) > 0 ) {
	$data->[0] = $dlf->[$self->{field_idx}];
	$data->[1] = $dlf;
    }
}

#------------------------------------------------------------------------
# Method merge_group_data( $dlf, $data )
#
# Method required by Lire::AsciiDlf::ReportOperator
sub merge_group_data {
    my ( $self, $value, $data ) = @_;

    return unless $value->{total} ne "";

    my $fields = [ split / /, $value->{total} ];

    unless (defined $data->[0]) {
	$data->[0] = $value->{value};
	$data->[1] = $fields;
	return;
    }

    # Change the value only if the fields sorts after the last one
    if ( $self->{sort_func}->( $fields, $data->[1] ) > 0 ) {
	$data->[0] = $value->{value};
	$data->[1] = $fields;
    }
}

#------------------------------------------------------------------------
# Method set_summary_value( $group, $data )
#
# Method required by Lire::AsciiDlf::GroupOp.
# We have to take care of non-numerical values
sub set_summary_value {
    my ( $self, $group, $data ) = @_;

    my $value = $data->[0];
    my $type	 = $self->{field_type} || "number";
    my $total = $self->{total_func}->( $data->[1] );
    my $content;
    if ( is_numeric_type( $self->{field_type} ) ) {
	$content = format_numeric_type( $value, $type );
    } else {
	$content = $value;
    }

    $group->set_summary_value( $self->name,
                               content => $content,
                               value => $value,
                               total => $total,
                               missing_cases => $self->missing_cases( $data ),
                             );
}

#------------------------------------------------------------------------
# Method add_entry_value( $entry, $data )
#
# Method required by Lire::AsciiDlf::ReportOperator
# We have to take care of non-numerical values
sub add_entry_value {
    my ( $self, $entry, $data ) = @_;

    my $value = $data->[0];
    my $total = $self->{total_func}->( $data->[1] );
    my $content;
    if ( is_numeric_type( $self->{field_type} ) ) {
	$content = format_numeric_type( $value, $self->{field_type} );
    } else {
	$content = $value;
    }

    $entry->add_value( content => $content,
                       value => $value,
                       total => $total,
                       missing_cases => $self->missing_cases( $data ),
                     );
}

# keep perl happy
1;

__END__

=pod

=head1 NAME

Lire::AsciiDlf::GroupOp -

=head1 SYNOPSIS


=head1 DESCRIPTION

Reports consist of several periods.  E.g. a weekly report can consist of 2002
week no 23, 2002 week no 24, 2002 week no 25, ..., 2002 week no 52, 2003 week
no 1, 2003 week no 2.

A Timegroup is a hash, which holds an array, which holds a report's data.  (We
omit the complication caused by the fact these things can be nested.)

See Lire::NestableAggregator:

    my $self = bless {
                      report_spec   => $report_spec,
                      element_name  => $args{element_name},
                      ops           => [],
                      op            => $args{element_name},
                      name          => $name,
                      field         ...
                      ...
                     }, $class;

Each period's data is contained in a slot in that array in the Timegroup hash.
E.g. week 2002 no 25 is represented by a certain slot.
And, if we have a 1 month period, e.g. the array is structured like

  [ sep2002, oct2002, dec2002, jan2003, ...]

Such an array is created by the init_group_data() routine.

All operations (like avg, count, sum, etc.) have an init_group_data(),
update_group_data() and end_group_data() method.  These 3 functions are
the high-level interface to GroupOp.

Suppose we're creating a monthly requests-by-day report, containing

 <timegroup><count/></timegroup>

.  In the data structure

 ([ sep2002, oct2002, ...])

sep2002 is the data structure created by the GroupOp::Count operation and which
is used to "count" the sep2002 stats.

If the report was a req-by-user-by-day period, sep2002 would be the data
structure created by the Group::init_group_data() method.

When not optimizing for large data sets, the group's data structure is a hash
instead of an array.  Keys of these hashes are the group's keys.  In such a
case, the values of this hash contain whatever data structure created by the
operator contained in the group element (count in the case of the last
example).

So, the structure boils down to:

 Timegroup: [ sep2002 = { user1 => ?, user2 => },
         oct2002 = { user1 => ?, user2 => }, etc. ]

Each { user1 => ?, user2 => } hash is created by the same Group instance.  The
update_group_data() function will be called with the appropriate data structure
for the DLF record we are currently processing


A NestableAggregator is an Aggregator that implements the GroupOp interface.
The only non-nestable Aggregator is the Summary element which doesn't do real
aggregation (all the DLF records are aggregated in one "group") whereas the
other aggregators (group, timeslot, timegroup, rangegroup) will aggregate the
DLF records into different groups (based on key (for Group), time period (for
Timegroup), etc.)


=head1 VERSION

$Id: GroupOp.pm,v 1.39 2003/05/16 21:43:38 flacoste Exp $

=head1 COPYRIGHT

Copyright (C) 2001, 2002 Stichting LogReport Foundation LogReport@LogReport.org

This file is part of Lire.

Lire is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program (see COPYING); if not, check with
http://www.gnu.org/copyleft/gpl.html or write to the Free Software 
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.

=head1 AUTHOR

Francis J. Lacoste <flacoste@logreport.org>

=cut
