package Lire::AsciiDlf::ReportOperator;

use strict;

use base qw/Exporter/;

use vars qw/ $VERSION @EXPORT_OK /;

use Exporter;
use Carp;

BEGIN {
    ($VERSION)	= '$Revision: 1.8 $' =~ m!Revision: ([.\d]+)!;
    @EXPORT_OK = qw/group_data_value/;
}

=pod

=head1 NAME

Lire::AsciiDlf::ReportOperator - Base interface for AsciiDlf operator implementation

=head1 SYNOPSIS

    # This module is only an interface description

=head1 DESCRIPTION

The Lire::AsciiDlf::ReportOperator module describes the method that
must be implemented by operator that can be used to compute Lire
report from Ascii DFL file.

It also defines the methods needed for the operator to properly
support merging of XML reports.

This interface only defines the method used by all operator that are
nested into an aggregator. The Lire::AsciiDlf::Aggregator modules
defines other methods that will need to be implemented by aggregator
operator.

=head1 REPORT COMPUTATION INTERFACE

=head2 init_report( $period_start, $period_end )

This method is called once before any DLF records is seen by the
operator. In this method, the operator may initialize any data
structure that will needed when the report.

A good thing to do in this method is to cache the index of the
DlfFields.

The $period_start and $period_end parameters gives the time spanned by
the DLF file.

=cut

sub init_report {
    croak "Unimplemented init_report() method in ", ref $_[0], "\n";
}

=pod

=head2 dlf_sort_fields()

This method will be called once the operator was inited to obtain the
sort order in which this operator needs to see the DLF records. If no
particular sort order is needed, it should returns an empty array.
Otherwise it should returns in an array the DLF field's names
according to which the records should be sorted.

This list cannot contains any numeric fields.

There is currently no way to specify a reverse sort order.

=cut

sub dlf_sort_fields {
    return ();
}


=pod

=head2 init_merge( $period_start, $period_end )

This method is called once before merging of XML reports is started.

The $period_start and $period_end parameters gives the time spanned by
all the XML reports.

=cut

sub init_merge {
    croak "Unimplemented init_merge() method in ", ref $_[0], "\n";
}

=pod

=head2 end_report()

This method is called once after all the DLF records were processed.

In this method, the operator might do the computation that can only be
done once all the records were processed.

This method is called before the operator is asked to fill the group's
entries.

=cut

sub end_report {
    croak "Unimplemented end_report() method in ", ref $_[0], "\n";
}

=pod

=head2 end_merge()

This method is once after all the XML reports were merged.

=cut

sub end_merge {
    croak "Unimplemented end_merge() method in ", ref $_[0], "\n";
}

=pod

=head2 init_group_data()

This method should return a data structure that will be passed the
call to update_group_data() and merge_group_data(). There will be one
such data structure for every group that exists in the aggregator
containing this operator.

There are a few restrictions placed on this data structure in order to
assure that it can be sorted by aggregator that need to sort their
entries. If the data structure is

=over

=item SCALAR

That value can be used for sorting purpose.

=item SCALAR REFERENCE

The scalar value will be used as sorting key.

=item ARRAY REFERENCE

The first item in the array reference will be used as sorting key.

=item HASH REFERENCE

The C<sort_value> items in the hash will be used as sorting key.

=item OBJECT REFERENCE

The value returned by the sort_value() method will be used as sorting
key.

=back

A function group_data_value() can be used to extract the sorting key
value according to these rules. It can be imported:

    use Lire::AsciiDlf::ReportOperator qw/group_data_value/;

    my $v = group_data_value( $data );

=cut

sub init_group_data {
    croak "Unimplemented init_group_data() method in ", ref $_[0], "\n";
}

=pod

=head2 update_group_data( $dlf, $data )

This method will be called once for every DLF records that should be
processed by the current operator. The $dlf parameter is an array
reference to the DLF fields of the record. The order of the DLF fields
in the array should be obtained by querying the Lire::DlfSchema
corresponding to the current report specification schema.

The $data parameter is the data structure that was created by the
init_group_data() of this operator. Its the data structure associated
with the group associated to $dlf.

=cut

sub update_group_data {
    croak "Unimplemented update_group_data() method", ref $_[0],"\n";
}

=pod

=head2 merge_group_data( $value, $data )

This method will be called once for each value generated by this
operator in the XML reports to be merged. 

The $value parameter can be either an hash reference or a
Lire::Report::Group object. It will be the same things that was
generated by this operator.

$data is the group data structure returned by init_group_data() for
the group where this $value should be merged.

=head2 end_group_data( $data )

This method will be called once for every group that was created by
the aggregator containing this operator. The $data parameter is the
data structure that was returned by init_group_data() for the current
group.

One possible use of this method would be, for example, for the avg
operator to compute the actual average since it cannot be done until
all the records are processed.

=cut

sub end_group_data {
    croak "Unimplemented end_group_data() method in", ref $_[0], "\n";
}

=pod

=head2 add_entry_value( $entry, $data )

This method will be called once for each group created in by the
aggregator. In this method, the operator should add the appropriate
name and value for this entry. $entry is the Lire::Report::Entry
object to which the names and/or values should be added. $data is the
data structure returned by init_group_data() corresponding for the
current group.

For example, its in this method that the sum operator will add the
value containing the sum for the current entry.

=cut

sub add_entry_value {
    croak "Unimplemented add_entry_value() method in ", ref $_[0], "\n";
}

=pod

=head1 UTILITY FUNCTION

=head2 group_data_value($data)

Returns the "value" from one data item created by a init_group_data
method.

=cut

sub group_data_value {
    if ( ref $_[0] eq 'SCALAR' ) {
	return ${$_[0]};
    } elsif (ref $_[0] eq 'ARRAY' ) {
	return $_[0][0];
    } elsif (ref $_[0] eq 'HASH' ) {
	return $_[0]{sort_value};
    } elsif (ref $_[0]) {
	return $_[0]->sort_value;
    } else {
	return$_[0];
    }
}

# keep perl happy
1;

__END__

=head1 SEE ALSO

Lire::AsciiDlf::Aggregator(3pm), Lire::Ascii::DlfInfo(3pm)

=head1 VERSION

$Id: ReportOperator.pm,v 1.8 2003/02/02 21:27:16 flacoste Exp $

=head1 COPYRIGHT

Copyright (C) 2002 Stichting LogReport Foundation LogReport@LogReport.org

This file is part of Lire.

Lire is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program (see COPYING); if not, check with
http://www.gnu.org/copyleft/gpl.html or write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.

=head1 AUTHOR

Francis J. Lacoste <flacoste@logreport.org>

=cut


