package Lire::Config::IniFile;

use strict;

use Config::IniFiles;
use Data::Dumper;

sub write_ini_section {
    my ($conf, $fh, $stype, $sname) = @_;

    if(defined $sname) {
        print $fh "[$stype/$sname]\n";
    } else {
        print $fh "[$stype]\n";
    }

    my $v = new Lire::Config::IniVisitor(
        conf => $conf, file => $fh, section => $stype, name => $sname);
    $v->visit($conf->spec);

    print $fh "\n";
}

sub write_ini_file {
    my ($conf, $xml, $fh, $ini) = @_;

    my $cfg = $conf->get_config_file($xml);

    write_ini_section($cfg->global, $fh, 'global');

    foreach ($cfg->service_names) {
        write_ini_section($cfg->service($_), $fh, 'service', $_);
    }

    foreach ($cfg->template_names) {
        write_ini_section($cfg->template($_), $fh, 'template', $_);
    }

    foreach ($cfg->job_names) {
        write_ini_section($cfg->job($_), $fh, 'job', $_);
    }
}

sub compatible_kind {
	my $item = $_[0];

	return 1 if UNIVERSAL::isa($item, 'Lire::Config::Scalar');

	if(UNIVERSAL::isa($item, 'Lire::Config::List')) {
		# We can deal with lists but only if it's a simple list of scalars
		my $spec = $item->spec;
		return 0 unless UNIVERSAL::isa($item, 'Lire::Config::Spec::List');
        my @components = $spec->components;
		return 0 unless @components == 1;

		my $test = $components[0]->instance;
		return 1 if UNIVERSAL::isa($test, 'Lire::Config::Scalar');;
	}
	return 0;
}

sub read_ini_section {
    my ($conf, $ini, $stype, $sname) = @_;

	my $sect;

    if(defined $sname) {
        $sect = "$stype/$sname";
    } else {
        $sect = $stype;
    }

	$ini = $ini->{$sect};

	# FIXME: handle deleted sections

	foreach(values %{$conf->get_all}) {
		if(!(defined ($ini->{$_->name})) && compatible_kind($_)) {
            $_->set(undef);
print STDERR "Unsetting $sect/", $_->name, "\n";
        }
	}

    foreach(keys %$ini) {
       if($conf->spec->has_component($_)) {
           $conf->get($_)->set(defined $ini->{$_} ? $ini->{$_} : '');
print STDERR "Setting $sect/$_ to ", $ini->{$_}, "\n";
       }
    }
}


sub read_ini_file {
    my ($conf, $xml, $fh, $inifile) = @_;

    my %ini;
    tie %ini, 'Config::IniFiles', (
        -file => $inifile,
        -nocase => 1,
        -allowcontinue => 1,
    );
    print Dumper(\%ini);

    my $cfg = $conf->get_config_file($xml);

	# FIXME: handle added sections

    read_ini_section($cfg->global, \%ini, 'global');

    foreach ($cfg->service_names) {
        read_ini_section($cfg->service($_), \%ini, 'service', $_);
    }

    foreach ($cfg->template_names) {
        read_ini_section($cfg->template($_), \%ini, 'template', $_);
    }

    foreach ($cfg->job_names) {
        read_ini_section($cfg->job($_), \%ini, 'job', $_);
    }
}


package Lire::Config::IniVisitor;
use base qw(Lire::Config::SpecVisitor);

use strict;
use Data::Dumper;
use Lire::Config::SpecVisitor;

sub new {
    my $proto = shift;
    my $class = ref $proto || $proto;
    my $self = $class->SUPER::new(@_);

    $self->{stack} = [];
    $self->{names} = [];

    $self;
}

sub view {
    my ($val, $indent) = @_;

    $indent = '' unless defined $indent;

    if(!defined $val) {
        print "$indent(undef)\n";
    } elsif(ref $val eq 'HASH') {
        print "$indent\{\n";
        while(my ($k,$v) = each %{$val}) {
            view($v, "$indent$k => ");
        }
        print "$indent}\n";
    } elsif(ref $val eq 'ARRAY') {
        print "$indent\[\n";
        foreach my $e (@{$val}) {
            view($e, "  $indent");
        }
        print "$indent]\n";
    } elsif(!ref $val) {
        print "$indent'$val'\n";
    } else {
        print $indent.(ref $val)."\n";
    }
}

sub traverse {
    my ($sub, $spec, $conf) = @_;

    if(!defined $conf) {
        return undef;
    } elsif(ref $conf eq 'ARRAY') {
        return [map { traverse($sub, $spec, $_) } @{$conf}];
    } elsif(ref $conf eq 'HASH') {
        return traverse($sub, $spec, $conf->{$spec->name});
    } else {
        return $sub->($spec)
            foreach($conf);
    }
}

sub compound_start {
    my ($self, $spec) = @_;
    unshift @{$self->{stack}}, (traverse sub { $_->get_all }, $spec, $self->{conf});
}

sub compound_end {
    my ($self, $spec) = @_;
    shift @{$self->{stack}};
}

sub list_start {
    my ($self, $spec) = @_;
    unshift @{$self->{stack}}, (traverse sub { [ $_->elements ] }, $spec, $self->{stack}[0]);
    push @{$self->{names}}, $spec->name;
}

sub list_end {
    my ($self, $spec) = @_;
    shift @{$self->{stack}};
    pop @{$self->{names}};
}

sub print_scalar {
    my ($self, $val) = @_;

    return '' unless defined $val;

    if(ref $val eq 'HASH') {
        return undef;
    } elsif(ref $val eq 'ARRAY') {
        return join(':', map { $self->print_scalar($_) } @{$val});
    } else {
        return $val;
    }
}

sub scalar {
    my ($self, $spec) = @_;
    my $conf = traverse sub { $_->get() }, $spec, $self->{stack}[0];

    return unless $spec->has_context($self->{section});

    push @{$self->{names}}, $spec->name;

    my $fh = $self->{file};

    if(defined $conf) {
        my $val = $self->print_scalar($conf);
        print $fh $self->{names}[0], " = $val\n";
    } else {
        print $fh '# ', $self->{names}[0], " = \n";
    }

    pop @{$self->{names}};
}

sub select_start {
    my ($self, $spec) = @_;
    unshift @{$self->{stack}}, [];
}

sub option {
    my ($self, $spec) = @_;
    push @{$self->{stack}[0]}, $spec->name;
}

sub select_end {
    my ($self, $spec) = @_;
    my $opts = shift @{$self->{stack}};
    my $conf = traverse sub { $_->as_value(); }, $spec, $self->{stack}[0];

    return unless $spec->has_context($self->{section});

    push @{$self->{names}}, $spec->name;

    my $fh = $self->{file};

    if(defined $conf) {
        my $val = $self->print_scalar($conf);
        print $fh $self->{names}[0], " = $val\n";
    } else {
        print $fh '# ', $self->{names}[0], " = ", join('|', @{$opts}), "\n";
    }

    pop @{$self->{names}};
}

1;

__END__

=pod

=pod

=head1 NAME

Lire::Config::IniFile - Load and Save Lire config in .ini files

=head1 DESCRIPTION

This package contains functions to load (read_ini_file) and save
(write_ini_file) Lire Configuration files as .ini files.

=head2 read_ini_file($conf, $xml, $fh, $ini)
=head2 write_ini_file($conf, $xml, $fh, $ini)

The first argument is the configuration object that contains the file,
usually this is simply the global Config singleton instance.

$xml is the name of the configuration file you want to parse/unparse,
it must be present in the Config object that was passed as the first
argument.

$fh is a type glob to a file handle referring to the output .ini file.
$ini is the filename of that same file.

=head1 VERSION

$Id: IniFile.pm,v 1.8 2003/05/05 03:44:40 ubiq Exp $

=head1 COPYRIGHT

Copyright (C) 2003 Stichting LogReport Foundation LogReport@LogReport.org

This file is part of Lire.

Lire is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program (see COPYING); if not, check with
http://www.gnu.org/copyleft/gpl.html or write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.

=head1 AUTHOR

Wessel Dankers <wsl@logreport.org>

=cut


