package Lire::DlfStream;

use strict;

use vars qw/ $VERSION /;

use Lire::Logger;
use Lire::DlfSchema;
use Lire::DataTypes qw/is_numeric_type/;

use Lire::Utils qw/read_keyvalue_file write_keyvalue_file/;
use Symbol;

use Carp;

BEGIN {
    ($VERSION)	= '$Revision: 1.9 $' =~ m!Revision: ([.\d]+)!;
};

=pod

=head1 NAME

Lire::DlfStream - Interface to DLF data stream

=head1 SYNOPSIS

  use Lire::DlfStore;

  my $store = Lire::DlfStore->open( "mystore" );
  my $dlf_stream = $store->open_dlf_stream( "www", "r" );

  print "Data begins on ", scalar localtime $dlf_stream->start_time, "\n";
  print "Data ends on   ", scalar localtime $dlf_stream->end_time, "\n";
  while ( my $dlf = $dlf_stream->read_dlf ) {
    ...
  }

=head1 DESCRIPTION

This object encapsulates DLF stream.

=cut

#------------------------------------------------------------------------
# Constructor new( $store, $name, $mode )
#
# Used by Lire::DlfStore
sub new {
    my ( $pkg, $store, $name, $mode, $sort_spec ) = @_;

    croak "store should be an Lire::DlfStore instance (not $store)"
      unless UNIVERSAL::isa( $store, "Lire::DlfStore" );

    croak "invalid mode: $mode (should be either 'r' or 'w')"
      unless $mode eq 'r' || $mode eq 'w';

    croak "only 'r' mode stream can use a sort_spec"
      if $mode ne 'r' && $sort_spec;

    my $self = bless { _store    => $store,
                       _name     => $name,
                       _mode     => $mode,
                       _sort_spec => $sort_spec,
                     }, $pkg;

    $self->_load_dlf_meta;
    $self->_load_dlf_stats;

    my $file = $store->storage_file( "dlf", "$name.dlf" );
    $file = ">> $file" if $mode eq 'w';

    if ( $sort_spec ) {
        $self->_sort_stream();
    } else {
        my $file = $store->storage_file( "dlf", "$name.dlf" );
        $file = ">> $file" if $mode eq 'w';
        my $fh = gensym;
        open $fh, $file
          or croak "can't open DLF file $file: $!";
        $self->{_fh} = $fh;
    }

    return $self;
}

=pod

=head2 name

Returns the schema's name of the DlfStream.

=cut

sub name {
    $_[0]{_name};
}

=pod

=head2 mode()

Returns the mode in which the DlfStream was opened.

=cut

sub mode {
    $_[0]{_mode};
}

=pod

=head2 sort_spec()

Returns the sort specification that is set on the stream.

=cut

sub sort_spec {
    $_[0]{_sort_spec};
}

=pod

=head2 close()

This method should be called when the Lire::DlfStream isn't needed
anymore, otherwise the same DlfStream cannot be opened until then.

=cut

sub close {
    my $self = $_[0];

    return unless $self->{_fh};
    CORE::close $self->{_fh}
        or croak "error closing DLF stream: $!";
    delete $self->{_fh};

    $self->_save_dlf_stats
      if $self->{_mode} eq 'w';

    $self->{_store}->_stream_is_closed( $self->{_name} );
}

sub DESTROY {
    $_[0]->close;
}

=pod

=head2 nrecords()

Returs the number of DLF records in the stream.

=cut

sub nrecords {
    $_[0]{_nrecords};
}

=pod

=head2 start_time()

Returns the timestamp of the oldest DLF record in the stream in
seconds since the epoch.

=cut

sub start_time {
    $_[0]{_start_time};
}

=pod

=head2 end_time()

Returns the timestamp of the newest DLF record in the stream in
seconds since the epoch.

=cut

sub end_time {
    $_[0]{_end_time};
}

#------------------------------------------------------------------------
# Method _load_dlf_meta()
#
sub _load_dlf_meta {
    my $self = $_[0];

    # Get the information from the newest schema
    my $schema = Lire::DlfSchema::load_schema( $self->{_name} );
    $self->{_fields} = [ map { $_->name } $schema->fields ];
    $self->{_ts_field} = $schema->timestamp_field->name;

    my $meta_file = $self->{_store}->storage_file( "dlf",
                                                   "$self->{_name}.meta" );

    # Check for the new stream case
    if ( ! -f $meta_file  ) {
        my $meta = { Schema => $self->{_name},
                     Fields => join( " ", @{$self->{_fields}}),
                   };

        write_keyvalue_file( $meta_file, $meta );
    } else {
        my $meta = read_keyvalue_file( $meta_file );

        croak "missing Schema value in DLF meta file"
          unless exists $meta->{Schema};
        croak "incompatible schema $self->{_name} <> $meta->{Schema}"
          if $meta->{Schema} ne $self->{_name};

        croak "missing Fields value in DLF meta file"
          unless exists $meta->{Fields};
        my @fields = split / /, $meta->{Fields};

        $self->_migrate_dlf_schema( \@fields )
          if join(" ", @{$self->{_fields}}) ne join (" ", @fields);
    }
}


#------------------------------------------------------------------------
# Method _migrate_dlf_schema( $old_fields)
#
# Check if we need to do a schema migration. A migration will occur if
# the DLF Stream is opened in write mode. Otherwise the "migration"
# would be done only when the actual DLF data is read.
sub _migrate_dlf_schema {
    my ($self, $from_fields ) = @_;

    # In read mode, we only had to save the fields. The conversion is
    # done at read time.
    if ( $self->{_mode} eq 'r' ) {
        my $schema = Lire::DlfSchema::load_schema( $self->{_name} );
        $self->{_file_fields} = $from_fields;
        $self->{_pos_mapping} = [];

        my $i = 0;
        foreach my $f ( @$from_fields ) {
            if ( $schema->has_field( $f ) ) {
                $self->{_pos_mapping}[$i] = $schema->field( $f )->pos;
            }
            $i++;
        }
        return;
    }

    my $old_fh = gensym;
    my $new_fh = gensym;

    my $meta_file = $self->{_store}->storage_file( "dlf",
                                                   "$self->{_name}.meta");
    my $dlf_file = $self->{_store}->storage_file( "dlf",
                                                  "$self->{_name}.dlf");
    open $old_fh, $dlf_file
      or croak "can't open DLF file '$dlf_file': $!";
    open $new_fh, "> $dlf_file.new"
      or croak "can't create new DLF file '$dlf_file.new': $!";

    my $line;
    while ( defined ( $line = <$old_fh> ) ) {
        chomp $line;
        my $dlf = _from_dlf( $from_fields, $line );
        print {$new_fh} _to_dlf( $self->{_fields}, $dlf ), "\n";
    }

    CORE::close $old_fh;
    CORE::close $new_fh
      or croak "error closing '$dlf_file.new': $!";

    my $meta = { Schema => $self->{_name},
                 Fields => join( " ", @{$self->{_fields}}),
               };
    write_keyvalue_file( "$meta_file.new", $meta );

    rename "$meta_file.new", $meta_file
      or croak "can't replace old DLF meta file with new one: $!";

    rename "$dlf_file.new", $dlf_file
      or croak "can't replace old DLF file with new one: $!";
}

# ------------------------------------------------------------------------
# Method _load_dlf_stats()
#
# We keep the stats in a separate file than the schema information
# because the stats can be lost without much problem, whereas
# recovering automatically from losing information in the .meta
# file isn't possible.
sub _load_dlf_stats {
    my $self = $_[0];

    my $stats_file = $self->{_store}->storage_file( "dlf",
                                                    "$self->{_name}.stats" );
    my $dlf_file = $self->{_store}->storage_file( "dlf",
                                                   "$self->{_name}.dlf" );
    $self->{_nrecords} = 0;
    $self->{_start_time} = undef;
    $self->{_end_time} = undef;

    return unless -f $dlf_file;

    my $need_rebuild = 1;
    if ( -f $stats_file ) {
        # Make sure that the stats file contains up-to-date
        # information.
        if ( -M $stats_file >= -M $dlf_file ) {
            my $stats = read_keyvalue_file( $stats_file );

            $self->{_nrecords} = $stats->{Records}
              if defined $stats->{Records} && $stats->{Records} =~ /^\d+$/;

            $self->{_start_time} = $stats->{'Start-Time'}
              if defined $stats->{'Start-Time'} && $stats->{'Start-Time'} =~ /^\d+$/;

            $self->{_end_time} = $stats->{'End-Time'}
              if defined $stats->{'End-Time'} && $stats->{'End-Time'} =~ /^\d+$/;

            $need_rebuild = 0
              if ( defined $self->{_nrecords} && 
                   defined $self->{_start_time} &&
                   defined $self->{_end_time} &&
                   $self->{_end_time} >= $self->{_start_time} );
        }
    }

    $self->_compute_dlf_stats
      if $need_rebuild;
}

#------------------------------------------------------------------------
# Method _save_dlf_stats()
#
# Reverse of _load_dlf_stats()
sub _save_dlf_stats {
    my $self =$_[0];

    my %stats =
      (
       Records => $self->{_nrecords},
       'Start-Time' => $self->{_start_time},
       'End-Time' => $self->{_end_time},
      );

    my $stats_file = $self->{_store}->storage_file( "dlf",
                                                    "$self->{_name}.stats" );
    write_keyvalue_file( "$stats_file.new", \%stats );
    rename "$stats_file.new", $stats_file
      or croak "can't replace DLF stats file by newer one: $!";
}

#------------------------------------------------------------------------
# Method _compute_dlf_stats()
# 
# When the DLF
sub _compute_dlf_stats {
    my $self = $_[0];

    my $fh = gensym;

    $self->{_nrecords} = 0;
    $self->{_start_time} = undef;
    $self->{_end_time} = undef;

    my $dlf_file = $self->{_store}->storage_file( "dlf",
                                                  "$self->{_name}.dlf" );
    open $fh, $dlf_file
      or croak "can't open DLF file '$dlf_file': $!";

    my $line;
    my $first = 1;
    while ( defined ( $line = <$fh> ) ) {
        chomp $line;
        my $dlf = _from_dlf( $self->{_fields}, $line );

        $self->{_nrecords}++;

        my $time = $dlf->{$self->{_ts_field}};
        $self->{_start_time} ||= $time;
        $self->{_end_time}   ||= $time;

        if ( defined $time && $time ne "" ) {
            $self->{_start_time} = $time
              if $time < $self->{_start_time};
            $self->{_end_time} = $time
              if $time > $self->{_end_time};
        }
    }
    CORE::close $fh;

    $self->_save_dlf_stats;
}

#------------------------------------------------------------------------
# Function _from_dlf( $fields, $line )
#
# Converts the $line which should contains the DLF fields defined in
# $fields as an hash reference.
sub _from_dlf {
    my ( $fields, $line ) = @_;

    my @fields = ();

    my $i=0;
    my $r = {};
    foreach my $f ( split /\t/, $line ) {
        $r->{$fields->[$i++]} = $f eq 'LIRE_NA' ? undef : $f;
    }
    croak "expected ", scalar @$fields, " fields, not ", $i
      if $i != @$fields;

    return $r;
}

#------------------------------------------------------------------------
# Function _to_dlf( $fields, $r )
#
# Encode the DLF hash reference $r for DLF storage
sub _to_dlf {
    my ( $fields, $r ) = @_;

    my $i = 0;
    my @f = ();
    foreach my $f ( @$fields ) {
        $f[$i] = defined $r->{$f} ? $r->{$f} : "LIRE_NA";
        #   replace 8 bit chars with ?
        #   replace control chars with ?
        $f[$i] =~ tr/\200-\377\000-\037/?/;
        $f[$i] =~ s/\t/        /g;
        $i++;
    }

    join( "\t", @f );
}

=pod

=head2 read_dlf()

Returns an hash reference containing the next DLF record in the
stream. It returns undef once the end of the stream is reached.

This method will throw an exception if the DlfStream isn't open in 'r'
mode or if there is an error reading the DLF record.

=cut
sub read_dlf {
    my $self = $_[0];

    croak "read_dlf() can't be called on a DlfStream open in 'w' mode"
      if $self->{_mode} ne 'r';

    my $line = readline $self->{_fh};
    return undef unless defined $line;

    chomp $line;
    if ( $self->{_file_fields} ) {
        # Handle schema conversion here
        my $r = _from_dlf( $self->{_file_fields}, $line );
        return { map { $_ => $r->{$_} } @{$self->{_fields}} };
    } else {
        return _from_dlf( $self->{_fields}, $line );
    }
}

=pod

=head2 read_dlf_aref()

Returns the next DLF record in the stream as an array reference. The
fields are in the order specified by the schema.

This method will throw an exception if the DlfStream isn't open in 'r'
mode or if there is an error reading the DLF record.

=cut
sub read_dlf_aref {
    my $self = $_[0];

    croak "read_dlf_aref() can't be called on a DlfStream open in 'w' mode"
      if $self->{_mode} ne 'r';

    my $line = readline $self->{_fh};
    return undef unless defined $line;

    chomp $line;

    my $count = $self->{_file_fields} ? @{$self->{_file_fields}} : @{$self->{_fields}};
    my @dlf = map { $_ eq 'LIRE_NA' ? undef : $_ } split /\t/, $line;

    croak "expected ", $count, " fields, not ", scalar @dlf
          if @dlf != $count;

    if ( $self->{_pos_mapping} ) {
        # Schema conversion needed
        my @new = ();

        # Grow the array
        $new[$#{$self->{_fields}}] = undef;

        my $i = 0;
        my $mapping = $self->{_pos_mapping};
        foreach my $f ( @dlf ) {
            if ( defined $mapping->[$i] ) {
                $new[$mapping->[$i]] = $f;
            }
            $i++;
        }

        return \@new;
    } else {
        return \@dlf;
    }
}

=pod

=head2 write_dlf( $r )

Writes the fields contained in the hash ref $r to the DLF stream.

This method will throw an exception if there is an error writing the
DLF record or if the stream isn't opened in 'w' mode.

=cut

sub write_dlf {
    my ( $self, $r ) = @_;

    croak "write_dlf() can't be called on a DlfStream open in 'r' mode"
      if $self->{_mode} ne 'w';

    croak "missing DLF hash reference"
      unless defined $r;

    croak "DLF record isn't an hash reference: $r"
      unless ref $r eq 'HASH';

    # Update stats
    $self->{_nrecords}++;
    my $time = $r->{$self->{_ts_field}};
    if ( defined $time && $time ne "" ) {
        unless ( defined $self->{_start_time} ) {
            $self->{_start_time} = $time;
            $self->{_end_time} = $time;
        }
        $self->{_start_time} = $time
          if $time < $self->{_start_time};
        $self->{_end_time} = $time
          if $time > $self->{_end_time};
    }

    print {$self->{_fh}} _to_dlf( $self->{_fields}, $r ), "\n";
}

#------------------------------------------------------------------------
# Method _sort_stream()
#
# Builds the sort command that is used to sort the file
# and to read from the stream.
sub _sort_stream {
    my $self = $_[0];

    my $schema = Lire::DlfSchema::load_schema( $self->{_name} );

    my @sfields = split /\s+/, $self->sort_spec;

    # Build sort(1) keys
    my @keys = ();
    foreach my $f ( @sfields ) {
        my $opts = "-";
        # Check for reverse option
        if ( substr( $f, 0, 1) eq '-' ) {
            $opts .= "r";
            $f = substr( $f, 1);
        }

        croak "field $f doesn't exists in $self->{_name} schema"
          unless $schema->has_field( $f );
        my $schema_field = $schema->field( $f );

        # Check for numeric type
        $opts .= "n"
          if is_numeric_type( $schema_field->type );

        # Determine field position
	my $pos;
        if ( $self->{_file_fields} ) {
            # Find the field's position on disk
            foreach (my $i=0; $i < @{$self->{_file_fields}}; $i++) {
                if ( $self->{_file_fields}[$i] eq $f ) {
                    $pos = $i;
                    last;
                }
            }
            next unless defined $pos;
        } else {
            $pos = $schema_field->pos
        }

        # sort(1) -k takes 1 indexed argument
        $pos += 1;
	push @keys, $opts . "k",  $pos . "," . $pos;
    }

    my $file = $self->{_store}->storage_file( "dlf", "$self->{_name}.dlf" );

    # No keys, simply open the DLF file
    unless ( @keys ) {
        $self->{_fh} = gensym;
        open {$self->{_fh}}, $file
          or croak "error opening '$file' DLF file: $!";
        return;
    }

    my $fh_r = gensym;
    my $fh_w = gensym;
    pipe $fh_r, $fh_w
      or croak "error creating pipe: $!";

    my $child_pid = fork;
    croak "error forking: $!" unless defined $child_pid;
    if ( $child_pid == 0 ) {
        CORE::close $fh_r
          or croak "error closing reading end of pipe: $!";

	# Clear the environment
	# This makes sure that we use POSIX C language sort order
	local %ENV = ( PATH => $ENV{PATH},
		       HOME => $ENV{HOME},
		       TMPDIR => defined $ENV{TMPDIR} ? $ENV{TMPDIR} : '',
		     );
	# Child
	open ( STDIN, $file )
	  or die "error redirecting stdin: $!\n" ;
	open ( STDOUT, ">&" . fileno $fh_w )
	  or die "error redirecting stdout: $!\n";
	exec "sort", "-t", "\t", @keys
	  or die "exec failed: $!\n";
    }
    CORE::close $fh_w
      or croak "error closing writing end of pipe: $!";

    $self->{_fh} = $fh_r;
}

1;
