package Lire::Records;

use strict;

use vars qw( $VERSION @ISA $RECORDS_ID );

use Lire::ReportOperator;

use Carp;

BEGIN {
    ($VERSION)	= '$Revision: 1.10 $' =~ m!Revision: ([.\d]+)!;
    @ISA = qw( Lire::ReportOperator );
    $RECORDS_ID = 0;
}

=pod

=head1 NAME

Lire::Records - Base class for implementation of the records operator.

=head1 SYNOPSIS

    use Lire::Records;

=head1 DESCRIPTION

This module is the base class for implementation of the records
operator. The records operator isn't an aggregator nor a group
operation. It only collects information on each DLF records which will
be included in the report.

=head1 CONSTRUCTOR

=head2 new( %params )

Creates a new instance of a records operator. In addition to the
normal report operator parameters, the records operator takes one
additional parameter:

=over

=item fields

A reference to an array containing the names of the DLF field that should be included in the report.

=back

=cut

sub new {
    my $proto = shift;
    my $class = ref( $proto) || $proto;

    my %args = @_;

    croak "missing fields parameter"
      unless defined $args{fields};
    unless ( defined $args{name}) {
	$args{name} = $args{report_spec}->create_opname( 'records' );
    }

    my $self = bless {}, $class;
    $self->SUPER::init( %args,
			op => 'records',
		      );

    $self->fields( $args{fields} );

    $self;
}

#------------------------------------------------------------------------
# Method print( $fh, $prefix )
#
# Implementaion of the method required by Lire::ReportOperator.
sub print {
    my ($self,$fh, $prefix) = @_;
    $fh	    ||= \*STDOUT;
    $prefix ||= 0;

    my $pfx = " " x $prefix;

    print $fh $pfx, '<lire:records fields="', join( " ", @{$self->{fields}}),
      qq{"/>\n};
}

=pod

=head1 METHOD

=head2 fields( [$new_fields] )

Returns a reference to an array containing the DLF field names that
will be included in the report.

You can use the $new_fields parameter to change that value.

=cut

sub fields {
    my ($self, $fields) = @_;

    if ( @_ == 2 ) {
	if ( defined $fields ) {
	    croak "$fields isn't an array reference"
	      unless UNIVERSAL::isa( $fields, 'ARRAY' );
	    die "fields cannot be empty\n"
	      unless @$fields;

	    foreach my $f ( @$fields ) {
		croak "$f isn't a defined field in the ",
		  $self->report_spec->schema->id, " schema"
		    unless $self->report_spec->schema->has_field( $f );
	    }
	} else {
	    croak "undefined fields\n";
	}
	$self->{fields} = $fields;
    }
    $self->{fields};
}

# ------------------------------------------------------------------------
# Method fill_group_info( $info )
#
# Implementation of the method required by Lire::ReportOperator
sub fill_group_info {
    my ( $self, $info ) = @_;

    my $group_info = $info->create_group_info( $self->name );
    foreach my $field ( @{$self->fields} ) {
	my $dlf_field = $self->report_spec->schema->field( $field );
	$group_info->create_column_info( $field, 'categorical',
					 $dlf_field->type,
					 $dlf_field->label, );
    }
}

# keep perl happy
1;

__END__

=head1 SEE ALSO

Lire::ReportSpec(3pm), Lire::GroupOp(3pm), Lire::ReportOperator(3pm),
Lire::Group(3pm), Lire::Timegroup(3pm), Lire::Timeslot(3pm),
Lire::Rangegroup(3pm)

=head1 VERSION

$Id: Records.pm,v 1.10 2003/02/07 23:27:21 flacoste Exp $

=head1 COPYRIGHT

Copyright (C) 2001 Stichting LogReport Foundation LogReport@LogReport.org

This file is part of Lire.

Lire is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program (see COPYING); if not, check with
http://www.gnu.org/copyleft/gpl.html or write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.

=head1 AUTHOR

Francis J. Lacoste <flacoste@logreport.org>

=cut

