package Lire::Timegroup;

use strict;

use vars qw( $VERSION @ISA );

use Lire::Aggregator;
use Lire::DataTypes qw( :special :time );

use Carp;

BEGIN {
    ($VERSION)	= '$Revision: 1.16 $' =~ m!Revision: ([.\d]+)!;
    @ISA = qw( Lire::Aggregator );
}

=pod

=head1 NAME

Lire::Timegroup - Base class for implementation of the timegroup aggregator

=head1 SYNOPSIS

    use Lire::Timegroup;

=head1 DESCRIPTION

This module is the base class for implementation of the timegroup
aggregator. This aggregator will split the DLF records based on a time
period controlled throught the period attribute. For example, using 1d
as the period value, this aggregator will group all records in the
same day period together.

=head1 CONSTRUCTOR

=head2 new( %params )

Creates a new instance of a timegroup aggregator. In addition to the
normal report operator parameters, the timegroup aggregator can take
several parameters:

=over

=item field

This optional parameter contains the DLF field which contains the time
value used to group the DLF records together. See the field() method
for more information.

=item period

This mandatory parameter should contains the period's length that will
be used to group the records. See the period() method for more
information.

=back

=cut

sub new {
    my $proto = shift;
    my $class = ref( $proto) || $proto;

    my %params = @_;

    croak( "Missing period parameter" )
      unless exists $params{period};

    my $self = bless {}, $class;
    $self->SUPER::init( %params, op => "timegroup" );

    $self->field( $params{field} );
    $self->period( $params{period} );

    $self;
}

=pod

=head1 METHODS

=head2 field( [$new_field] )

Returns the DLF field's name that is used to group the DLF records.
This should be a valid timestamp DLF field in the current schema. By
default, the default timestamp field of the DLF schema is used.

You can change the field by passing a $new_field parameter.

=cut

sub field {
    my ( $self, $field ) = @_;

    if (@_ == 2 ) {
	if ( defined $field ) {
	    croak "$field isn't a valid field for the ",
	      $self->report_spec->schema->id, " schema"
		unless $self->report_spec->schema->has_field( $field );

	    croak "$field isn't a timestamp field"
	      unless $self->report_spec->schema->field( $field )->type()
		eq "timestamp";

	} else {
	    $field = $self->report_spec->schema->timestamp_field->name();
	}
	$self->{field} = $field;
    }

    return $self->{field};
}

=pod

=head2 period( [$new_period])

Returns the period's length in which the records are grouped. This can
either be a duration value or the name of a report specification's
parameter containing a duration value.

The period's length can be changed by using the $new_period parameter.

=cut

sub period {
    my ( $self, $period ) = @_;

    if (@_ == 2 ) {
	if ( $period =~ /^\$/ ) {
	    my $name = substr $period, 1;
	    croak "$name isn't a defined parameter"
	      unless $self->report_spec->has_param( $name );

	    croak "$name parameter isn't of type duration"
	      unless $self->report_spec->param( $name )->type()
		eq "duration";
	} else {
	    croak "$period isn't a valid duration"
	      unless check_duration( $period );
	}

	$self->{period} = $period;
    }

    return $self->{period};
}

# ------------------------------------------------------------------------
# Method create_categorical_info( $info )
#
# Implementation of the method required by Lire::Aggregator
sub create_categorical_info {
    my ( $self, $info ) = @_;

    my $dlf_field = $self->report_spec->schema->field( $self->field );
    $info->create_column_info( $self->field, 'categorical',
			       $dlf_field->type, $self->label );
}

# ------------------------------------------------------------------------
# Method xml_attrs( $info )
#
# Implementation of the method required by Lire::Aggregator
sub xml_attrs {
    my ( $self ) = @_;

    return qq{field="$self->{field}" period="$self->{period}"};
}

# keep perl happy
1;

__END__

=head1 SEE ALSO

Lire::ReportSpec(3pm), Lire::Group(3pm), Lire::ReportOperator(3pm),
Lire::Timeslot(3pm)

=head1 VERSION

$Id: Timegroup.pm,v 1.16 2003/02/07 23:27:23 flacoste Exp $

=head1 COPYRIGHT

Copyright (C) 2001-2002 Stichting LogReport Foundation LogReport@LogReport.org

This file is part of Lire.

Lire is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program (see COPYING); if not, check with
http://www.gnu.org/copyleft/gpl.html or write to the Free Software 
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.

=head1 AUTHOR

Francis J. Lacoste <flacoste@logreport.org>

=cut
