package tests::AsciiDlfFilterExprTest;

use strict;

use base qw/Lire::Test::TestCase/;

use Lire::ReportSpec;
use Lire::Param;
use Lire::AsciiDlf::FilterExpr;

use File::Basename qw/dirname/;
use Cwd qw/realpath/;
use Carp;

use constant TRUE => 1;
use constant FALSE => 0;

sub new {
    my $self = shift->SUPER::new( @_ );

    $self->{schemasdir} = realpath( dirname(__FILE__) . "/schemas" );

    $self->{report_spec} =  new Lire::ReportSpec();
    $self->{report_spec}->id( 'test-filter' );
    $self->{report_spec}->superservice( 'test' );
    $self->{report_spec}->title( 'Test FilterExpr Report Specification' );
    $self->{report_spec}->display_title( 'Test FilterExpr Report' );

    $self->{report_spec}->param( "num_param", 
                                 new Lire::Param( name => "num_param",
                                                  type => "int" ));
    $self->{report_spec}->param( "str_param", 
                                 new Lire::Param( name => "str_param",
                                                  type => "string" ));

    $self;
}

sub set_up {
    my $self = shift->SUPER::set_up();

    $self->{cfg}{lr_schemas_path} = [ $self->{schemasdir} ];
}

sub dlf_hash2array {
    my ($self, $dlf_hash) = @_;

    my $schema = Lire::DlfSchema::load_schema( "test" );
    my @dlf = ();
    while (my ($key,$value) = each %$dlf_hash ) {
        $dlf[$schema->field( $key )->pos] = $value;
    }
    \@dlf;
}

sub run_filter_test {
    my ($self, $template ) = @_;

    foreach my $name ( qw/op new_params params_init results_for_dlf/ ) {
        croak "missing key '$name' in filter test template"
          unless defined $template->{$name};
    }

    my $op;
    {
        no strict 'refs';

        my $pkg = 'Lire::AsciiDlf::FilterExpr::' . $template->{op};

        $op = $pkg->new( container => $self->{report_spec},
                         %{$template->{new_params}} );
        $self->assert_not_null( $op, "$pkg->new() returned undef" );
    }

    while ( my( $name, $value ) = each %{$template->{params_init}}) {
        $self->{report_spec}->param( $name )->value( $value );
    }

    my $filter_func = $op->compile;
    $self->assert_equals( 'CODE', ref $filter_func,
                          "compile() didn't return a CODE ref" );

    my @tests = @{$template->{results_for_dlf}};
    while ( @tests ) {
        my $result = shift @tests;
        my $dlf_h  = shift @tests;
        my $dlf    = $self->dlf_hash2array( $dlf_h );

        my $msg = $self->build_msg_failure( $template, $result, $dlf_h );
        if ( $result ) {
            $self->assert( $filter_func->( $dlf ), $msg );
        } else {
            $self->assert( !$filter_func->( $dlf ), $msg );
        }
    }
}

sub build_msg_failure {
    my ( $self, $template, $result, $dlf ) = @_;

    my $msg = "$template->{op} should have returned " .
      ( $result ? 'true' : 'false' ) . ":\n    new_params => { ";
    foreach my $key ( keys %{$template->{new_params}} ) {
        $msg .= $key . ' => ' . quote_value($template->{new_params}{$key}) . ', ';
    }
    $msg .= "}\n    params_init => { ";
    foreach my $key ( keys %{$template->{params_init}} ) {
        $msg .= $key . ' => ' . quote_value($template->{params_init}{$key}) . ', ';
    }

    $msg .= "}\n    dlf => { ";
    foreach my $key ( keys %$dlf ) {
        $msg .= $key . ' => ' . quote_value( $dlf->{$key} );
    }

    $msg .= "} ";
    return $msg;
}

sub quote_value {
    defined $_[0] ? "'$_[0]'" : 'undef';
}

sub test_gt {
    my $self = $_[0];

    $self->run_filter_test( {
                             op => 'Gt',
                             new_params => { arg1 => '$file_size',
                                             arg2 =>'$num_param' },
                             params_init => { num_param => 2048 },
                             results_for_dlf => [ TRUE, { file_size => 51_023},
                                                  FALSE, { file_size => 512 },
                                                  FALSE, { file_size => 2048 },
                                                  FALSE, { file_size => undef },
                                                ],
                            },
                          );

    $self->run_filter_test( {
                             op => 'Gt',
                             new_params => { arg1 => '$file_size',
                                             arg2 =>'$num_param' },
                             params_init => { num_param => undef },
                             results_for_dlf => [ FALSE, { file_size => 51_023},
                                                  FALSE, { file_size => undef },
                                                ],
                            },
                          );
}

sub test_ge {
    my $self = $_[0];

    $self->run_filter_test( {
                             op => 'Ge',
                             new_params => { arg1 => '$file_size',
                                             arg2 =>'$num_param' },
                             params_init => { num_param => 2048 },
                             results_for_dlf => [ TRUE, { file_size => 51_023},
                                                  FALSE, { file_size => 512 },
                                                  TRUE, { file_size => 2048 },
                                                  FALSE, { file_size => undef },
                                                ],
                            },
                          );

    $self->run_filter_test( {
                             op => 'Ge',
                             new_params => { arg1 => '$file_size',
                                             arg2 =>'$num_param' },
                             params_init => { num_param => undef },
                             results_for_dlf => [ FALSE, { file_size => 51_023},
                                                  FALSE, { file_size => undef },
                                                ],
                            },
                          );
}

sub test_lt {
    my $self = $_[0];

    $self->run_filter_test( {
                             op => 'Lt',
                             new_params => { arg1 => '$file_size',
                                             arg2 =>'$num_param' },
                             params_init => { num_param => 2048 },
                             results_for_dlf => [ FALSE, { file_size => 51_023},
                                                  TRUE, { file_size => 512 },
                                                  FALSE, { file_size => 2048 },
                                                  FALSE, { file_size => undef },
                                                ],
                            },
                          );

    $self->run_filter_test( {
                             op => 'Lt',
                             new_params => { arg1 => '$file_size',
                                             arg2 =>'$num_param' },
                             params_init => { num_param => undef, },
                             results_for_dlf => [ FALSE, { file_size => 51_023},
                                                  FALSE, { file_size => undef },
                                                ],
                            },
                          );
}

sub test_le {
    my $self = $_[0];

    $self->run_filter_test( {
                             op => 'Le',
                             new_params => { arg1 => '$file_size',
                                             arg2 =>'$num_param' },
                             params_init => { num_param => 2048 },
                             results_for_dlf => [ FALSE, { file_size => 51_023},
                                                  TRUE, { file_size => 512 },
                                                  TRUE, { file_size => 2048 },
                                                  FALSE, { file_size => undef },
                                                ],
                            },
                          );

    $self->run_filter_test( {
                             op => 'Le',
                             new_params => { arg1 => '$file_size',
                                             arg2 =>'$num_param' },
                             params_init => { num_param => undef, },
                             results_for_dlf => [ FALSE, { file_size => 51_023},
                                                  FALSE, { file_size => undef },
                                                ],
                            },
                          );
}

sub test_eq {
    my $self = $_[0];

    $self->run_filter_test( {
                             op => 'Eq',
                             new_params => { arg1 => '$user',
                                             arg2 =>'$str_param' },
                             params_init => { str_param => 'flacoste' },
                             results_for_dlf => [ FALSE, { user => 'joostvb' },
                                                  TRUE, { user => 'flacoste' },
                                                  FALSE, { user => undef },
                                                ],
                            },
                          );

    $self->run_filter_test( {
                             op => 'Eq',
                             new_params => { arg1 => '$user',
                                             arg2 =>'$str_param' },
                             params_init => { str_param => undef, },
                             results_for_dlf => [
                                                 FALSE, { user => undef },
                                                ],
                            },
                          );
}

sub test_eq_num {
    my $self = $_[0];

    $self->run_filter_test( {
                             op => 'Eq',
                             new_params => { arg1 => '$time_taken',
                                             arg2 => 5 },
                             params_init => { },
                             results_for_dlf => [ TRUE, { time_taken => 5 },
                                                  FALSE, { time_taken => undef },
                                                  FALSE, { time_taken => 0 },
                                                ],
                            },
                          );
}

sub test_ne {
    my $self = $_[0];

    $self->run_filter_test( {
                             op => 'Ne',
                             new_params => { arg1 => '$user',
                                             arg2 =>'$str_param' },
                             params_init => { str_param => 'flacoste' },
                             results_for_dlf => [ TRUE, { user => 'joostvb' },
                                                  FALSE, { user => 'flacoste' },
                                                  FALSE, { user => undef },
                                                ],
                            },
                          );

    $self->run_filter_test( {
                             op => 'Ne',
                             new_params => { arg1 => '$user',
                                             arg2 =>'$str_param' },
                             params_init => { str_param => undef, },
                             results_for_dlf => [
                                                 FALSE, { user => undef },
                                                ],
                            },
                          );
}

sub test_ne_num {
    my $self = $_[0];

    $self->run_filter_test( {
                             op => 'Ne',
                             new_params => { arg1 => '$time_taken',
                                             arg2 => 5 },
                             params_init => { },
                             results_for_dlf => [ FALSE, { time_taken => 5 },
                                                  FALSE, { time_taken => undef },
                                                  TRUE, { time_taken => 0 },
                                                ],
                            },
                          );
}

sub test_match {
    my $self = $_[0];

    $self->run_filter_test( {
                             op => 'Match',
                             new_params => { value => '$user',
                                             re =>'$str_param',
                                             'case-sensitive' => 0,
                                           },
                             params_init => { str_param => '^F' },
                             results_for_dlf => [ FALSE, { user => 'joostvb' },
                                                  TRUE, { user => 'flacoste' },
                                                  FALSE, { user => undef },
                                                ],
                            },
                          );

    $self->run_filter_test( {
                             op => 'Match',
                             new_params => { value => '$user',
                                             re =>'$str_param',
                                             'case-sensitive' => 1,
                                           },
                             params_init => { str_param => '^f' },
                             results_for_dlf => [ FALSE, { user => 'Francis' },
                                                  TRUE, { user => 'flacoste' },
                                                  FALSE, { user => undef },
                                                ],
                            },
                          );

    eval {
        new Lire::AsciiDlf::FilterExpr::Match( container => $self->{report_spec},
                                               value => '$user',
                                               re => '$time_taken',
                                             );
    };
    $self->assert_not_null( $@, 're() with a ref to a DLF field should have failed' );
}

1;
