package tests::ReportMergerTest;

use strict;

use base qw/ Lire::Test::TestCase tests::TestCaseUsingOldConfigVars/;

use Lire::ReportSpec;
use Lire::FilterSpec;
use Lire::DlfSchema;
use Lire::ReportMerger;
use Lire::ReportParser::ReportBuilder;
use File::Basename qw/dirname/;
use Cwd qw/realpath/;
use Carp;

sub new {
    my $self = shift()->SUPER::new( @_ );

    $self->{dir} = realpath( dirname( __FILE__ ));
    $self->setup_old_config_vars;
    $self;
}

sub set_up {
    my $self = shift->SUPER::set_up();

    # This is the time zone in which the report with which we compare
    # was created.
    $self->{saved_tz} = $ENV{TZ};
    $ENV{TZ} = 'EST';
    $self->{cfg}{lr_reports_path} = [ $self->{dir} . "/reports" ];
    $self->{cfg}{lr_filters_path} = [ $self->{dir} . "/filters" ];
    $self->{cfg}{lr_schemas_path} = [ $self->{dir} . "/schemas" ];
}

sub tear_down {
    my $self = shift->SUPER::tear_down();

    # Perl has the bright idea of warning when
    # one assigns undef to an environment variable ?!?
    $ENV{TZ} = defined $self->{saved_tz} ? $self->{saved_tz} : '';
}

sub test_self_merge_report {
    my $self = $_[0];

    my $parser = new Lire::ReportParser::ReportBuilder;

    # We cannot use test.xml as a comparison because
    # there the relative ordering will change due to
    # the difference between sort(1) and perl's builtin sort.
    my $report_e = $parser->parsefile( "$self->{dir}/data/test-selfmerge.xml" );

    my $gen = new Lire::ReportMerger( "test",
                                      $self->{dir} . "/data/test.cfg",
                                      $self->{dir} . "/data/test.xml",
                                    );
    my $report_t = $gen->merge_reports;

    # The difference in these fields are normal
    $report_t->{date} = $report_e->{date};
    $report_t->generator( $report_e->generator );

    $self->assert_deep_equals( $report_e, $report_t );
}

sub test_doublemerge_report {
    my $self = $_[0];

    my $parser = new Lire::ReportParser::ReportBuilder;

    my $report_e = $parser->parsefile( "$self->{dir}/data/test-doublemerge.xml" );

    my $gen = new Lire::ReportMerger( "test",
                                      $self->{dir} . "/data/test.cfg",
                                      $self->{dir} . "/data/test.xml",
                                      $self->{dir} . "/data/test.xml",
                                    );
    my $report_t = $gen->merge_reports;

    # The difference in these fields are normal
    $report_t->{date} = $report_e->{date};
    $report_t->generator( $report_e->generator );

    $self->assert_deep_equals( $report_e, $report_t );
}

sub test_merge_with_empty_report {
    my $self = $_[0];

    my $warnings = "";
    local $SIG{__WARN__} = sub { $warnings .= join "", @_ };

    my $parser = new Lire::ReportParser::ReportBuilder;
    my $report_e = $parser->parsefile( "$self->{dir}/data/test.xml" );

    my $gen = new Lire::ReportMerger( "test",
                                      $self->{dir} . "/data/test.cfg",
                                      $self->{dir} . "/data/test.xml",
                                      $self->{dir} . "/data/test-empty.xml",
                                    );
    my $report_t = $gen->merge_reports;
    $self->annotate( $warnings );
    $self->assert( !$warnings, "There were warnings" );
    # The difference in these fields are normal
    $report_t->{date} = $report_e->{date};
    $report_t->generator( $report_e->generator );

    $self->assert_deep_equals( $report_e, $report_t );
}

sub test_two_empty_report {
    my $self = $_[0];

    my $warnings = "";
    local $SIG{__WARN__} = sub { $warnings .= join "", @_ };

    my $parser = new Lire::ReportParser::ReportBuilder;
    my $report_e = $parser->parsefile( "$self->{dir}/data/test-empty.xml" );

    my $gen = new Lire::ReportMerger( "test",
                                      $self->{dir} . "/data/test.cfg",
                                      $self->{dir} . "/data/test-empty.xml",
                                      $self->{dir} . "/data/test-empty.xml",
                                    );
    my $report_t = $gen->merge_reports;
    $self->annotate( $warnings );
    $self->assert( !$warnings, "There were warnings" );

    # The difference in these fields are normal
    $report_t->{date} = $report_e->{date};
    $report_t->generator( $report_e->generator );

    $self->assert_deep_equals( $report_e, $report_t );
}

1;
